@@ -33,8 +33,12 @@ import time
 
 DESCRIPTION_MAX = 1024
 
+logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')
 
-class GitHub2GitLab:
+log = logging.getLogger(__name__)
+
+
+class GitHub2GitLab(object):
     TAG_MERGED = ":MERGED:"
 
     STATE_EVENT2MERGE_STATE = {
@@ -79,8 +83,7 @@ class GitHub2GitLab:
         else:
             level = logging.INFO
 
-        logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
-                            level=level)
+        logging.getLogger('github2gitlab').setLevel(level)
 
         self.tmpdir = "/tmp"
 
@@ -111,6 +114,9 @@ class GitHub2GitLab:
         parser.add_argument('--ignore-closed', action='store_const',
                             const=True,
                             help='ignore pull requests closed and not merged')
+        parser.add_argument('--skip-pull-requests', action='store_const',
+                            const=True,
+                            help='do not mirror PR to MR')
         parser.add_argument('--verbose', action='store_const',
                             const=True,
                             help='enable verbose (debug) logging')
@@ -128,24 +134,39 @@ class GitHub2GitLab:
         if self.add_project():
             self.unprotect_branches()
         self.git_mirror()
-        self.pull_requests = self.get_pull_requests()
-        self.merge_requests = self.get_merge_requests()
-        self.update_merge_pull()
-        self.sync()
+        if not self.args.skip_pull_requests:
+            self.pull_requests = self.get_pull_requests()
+            self.merge_requests = self.get_merge_requests()
+            self.update_merge_pull()
+            self.sync()
         return 0
 
     def sh(self, command):
-        logging.debug(command)
-        output = ''
-        try:
-            output = subprocess.check_output(command, stderr=subprocess.STDOUT,
-                                             shell=True)
-        except subprocess.CalledProcessError as e:
-            output = e.output
-            raise e
-        finally:
-            logging.debug(command + " output " + output)
-        return output
+        log.debug(":sh: " + command)
+        proc = subprocess.Popen(
+            args=command,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.STDOUT,
+            shell=True,
+            bufsize=1)
+        lines = []
+        with proc.stdout:
+            for line in iter(proc.stdout.readline, b''):
+                line = line.decode('utf-8')
+                lines.append(line)
+                log.debug(str(line.strip()))
+        if proc.wait() != 0:
+            raise subprocess.CalledProcessError(
+                returncode=proc.returncode,
+                cmd=command
+            )
+        return "".join(lines)
+
+    def gitlab_create_remote(self, repo):
+        repo.create_remote('gitlab',
+                           self.gitlab['git'] + ":" +
+                           self.gitlab['namespace'] + "/" +
+                           self.gitlab['name'] + ".git")
 
     def git_mirror(self):
         name = self.gitlab['name']
@@ -155,28 +176,71 @@ class GitHub2GitLab:
         repo = git.Repo(name)
         os.chdir(name)
         if not hasattr(repo.remotes, 'gitlab'):
-            repo.create_remote('gitlab',
-                               self.gitlab['git'] + ":" +
-                               self.gitlab['namespace'] + "/" +
-                               self.gitlab['name'] + ".git")
-        self.sh("git fetch --prune --force origin +refs/heads/*:refs/heads/* "
-                "+refs/tags/*:refs/tags/*")
-        self.sh("git fetch --prune origin +refs/pull/*:refs/heads/pull/*")
+            self.gitlab_create_remote(repo)
         if 'branches' in self.github:
-            refs = []
-            for branch in self.github['branches']:
-                refs.append("+refs/heads/" + branch + ":" +
-                            "refs/heads/" + branch)
-            self.sh("git fetch --force origin " + " ".join(refs))
-            self.sh("git push --force gitlab " + " ".join(refs))
+            branches_ref = " ".join([
+                "+refs/heads/{b}:refs/heads/{b}".format(b=b)
+                for b in self.github['branches']
+            ])
         else:
-            self.sh("git push --prune gitlab +refs/heads/*:refs/heads/*")
-        self.sh("git push --prune --force gitlab "
-                "+refs/heads/pull/*:refs/heads/pull/* "
+            branches_ref = "+refs/heads/*:refs/heads/*"
+        #
+        # Fetch
+        #
+        self.sh("git fetch --force origin " +
+                branches_ref +
+                " +refs/tags/*:refs/tags/*")
+        #
+        # Track refs
+        #
+        if self.args.skip_pull_requests:
+            self.git_mirror_optimize(repo)
+        else:
+            self.sh("git fetch origin +refs/pull/*:refs/heads/pull/*")
+        #
+        # Push
+        #
+        self.sh("git push --prune --force gitlab " +
+                branches_ref + " " +
+                "+refs/heads/pull/*:refs/heads/pull/* " +
                 "+refs/tags/*:refs/tags/* ")
         os.chdir("..")
         self.revision2commit = {}
 
+    def git_mirror_optimize(self, repo):
+        self.sh("git fetch origin +refs/pull/*:refs/remotes/origin/pull/*")
+        for head in repo.refs:
+            pr = re.search('^origin/pull/(\d+)/head$', head.name)
+            if not pr:
+                continue
+            pr = pr.group(1)
+            merge_name = 'origin/pull/' + pr + '/merge'
+            if merge_name not in repo.refs:
+                log.debug(head.name + " cannot merge, ignore")
+                continue
+            merge = repo.commit(merge_name)
+            if merge.parents[1] != head.commit:
+                log.debug(head.name + " merge is obsolete, skip")
+                continue
+            known_head_name = 'pull/' + pr + '/head'
+            try:
+                known_head = repo.commit(known_head_name)
+                if known_head == head.commit:
+                    log.debug(head.name + " head has not moved, skip")
+                    continue
+                action = 'update'
+            except gitdb.exc.BadName:
+                action = 'create'
+            log.debug(action + " ref " + known_head_name + "==" +
+                      head.commit.hexsha)
+            repo.git.update_ref('refs/' + known_head_name, head.commit)
+            log.info(action + " branch " + 'pull/' + pr + "/merge == " +
+                     merge.hexsha)
+            repo.git.update_ref('refs/heads/pull/' + pr + '/merge', merge)
+            log.info(action + " branch " + 'pull/' + pr + "/head == " +
+                     head.commit.hexsha)
+            repo.git.update_ref('refs/heads/pull/' + pr + '/head', head.commit)
+
     def add_key(self):
         "Add ssh key to gitlab if necessary"
         with open(self.args.ssh_public_key) as f:
@@ -185,14 +249,14 @@ class GitHub2GitLab:
         url = g['url'] + "/user/keys"
         query = {'private_token': g['token']}
         keys = requests.get(url, params=query).json()
-        logging.debug("looking for '" + public_key + "' in " + str(keys))
+        log.debug("looking for '" + public_key + "' in " + str(keys))
         if (list(filter(lambda key: key['key'] == public_key, keys))):
-            logging.debug(self.args.ssh_public_key + " already exists")
+            log.debug(self.args.ssh_public_key + " already exists")
             return None
         else:
             name = 'github2gitlab'
-            logging.info("add " + name + " ssh public key from " +
-                         self.args.ssh_public_key)
+            log.info("add " + name + " ssh public key from " +
+                     self.args.ssh_public_key)
             query['title'] = name
             query['key'] = public_key
             result = requests.post(url, query)
@@ -206,10 +270,10 @@ class GitHub2GitLab:
         url = g['url'] + "/projects/" + g['repo']
         query = {'private_token': g['token']}
         if (requests.get(url, params=query).status_code == requests.codes.ok):
-            logging.debug("project " + url + " already exists")
+            log.debug("project " + url + " already exists")
             return None
         else:
-            logging.info("add project " + g['repo'])
+            log.info("add project " + g['repo'])
             url = g['url'] + "/projects"
             query['public'] = 'true'
             query['namespace'] = g['namespace']
@@ -217,8 +281,8 @@ class GitHub2GitLab:
             result = requests.post(url, params=query)
             if result.status_code != requests.codes.created:
                 raise ValueError(result.text)
-            logging.debug("project " + g['repo'] + " added: " +
-                          result.text)
+            log.debug("project " + g['repo'] + " added: " +
+                      result.text)
             return result.json()
 
     def unprotect_branches(self):
@@ -328,12 +392,12 @@ class GitHub2GitLab:
                 if updates:
                     self.update_merge_request(merge, updates)
                 else:
-                    logging.debug("https://github.com/" +
-                                  self.github['repo'] + "/" +
-                                  "pull/" + number + " == " +
-                                  self.gitlab['host'] + "/" +
-                                  parse.unquote(self.gitlab['repo']) + "/" +
-                                  "merge_requests/" + str(merge['iid']))
+                    log.debug("https://github.com/" +
+                              self.github['repo'] + "/" +
+                              "pull/" + number + " == " +
+                              self.gitlab['host'] + "/" +
+                              parse.unquote(self.gitlab['repo']) + "/" +
+                              "merge_requests/" + str(merge['iid']))
 
     def rev_parse(self, pull, revision):
         if revision in self.revision2commit:
@@ -344,10 +408,10 @@ class GitHub2GitLab:
                 repo.rev_parse("heads/" + revision)
                 return True
             except gitdb.exc.BadName:
-                logging.debug("ignore https://github.com/" +
-                              self.github['repo'] + "/pull/" +
-                              str(pull['number']) + " because " +
-                              revision + " is not a known revision")
+                log.debug("ignore https://github.com/" +
+                          self.github['repo'] + "/pull/" +
+                          str(pull['number']) + " because " +
+                          revision + " is not a known revision")
                 return False
 
     @staticmethod
@@ -356,7 +420,7 @@ class GitHub2GitLab:
         try:
             return json.loads(payload)
         except ValueError as e:
-            logging.error("unable to json.loads(" + payload + ")")
+            log.error("unable to json.loads(" + payload + ")")
             raise e
 
     def get(self, url, query, cache):
@@ -368,7 +432,7 @@ class GitHub2GitLab:
             payloads = []
             next_query = query
             while next_query:
-                logging.debug(str(next_query))
+                log.debug(str(next_query))
                 result = requests.get(url, params=next_query)
                 payloads += result.json()
                 next_query = None
@@ -422,12 +486,12 @@ class GitHub2GitLab:
         g = self.gitlab
         query['private_token'] = g['token']
         url = g['url'] + "/projects/" + g['repo'] + "/merge_requests"
-        logging.info('create_merge_request: ' + str(query))
+        log.info('create_merge_request: ' + str(query))
         result = requests.post(url, params=query)
         if result.status_code != requests.codes.created:
             raise ValueError(result.text)
         merge = result.json()
-        logging.debug('merge ' + str(merge))
+        log.debug('merge ' + str(merge))
         for (key, value) in six.iteritems(query):
             if key == 'private_token':
                 continue
@@ -454,7 +518,7 @@ class GitHub2GitLab:
         updates['private_token'] = g['token']
         url = (g['url'] + "/projects/" + g['repo'] + "/merge_request/" +
                str(merge_request['id']))
-        logging.info('update_merge_request: ' + url + ' <= ' + str(updates))
+        log.info('update_merge_request: ' + url + ' <= ' + str(updates))
         return requests.put(url, params=updates).json()
 
     def verify_merge_update(self, updates, result):
@@ -475,3 +539,7 @@ class GitHub2GitLab:
                                      key=key,
                                      value=value,
                                      result=result_value))
+
+# Local Variables:
+# compile-command: "cd .. ; virtualenv/bin/tox -e flake8"
+# End:
