@@ -17,13 +17,14 @@
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see `<http://www.gnu.org/licenses/>`.
 #
-import fixtures
-import json
+import git
+import gitdb
 import logging
 import mock
 import os
+import pytest
+import shutil
 import tempfile
-import testtools
 
 from github2gitlab import main
 
@@ -31,35 +32,35 @@ logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                     level=logging.DEBUG)
 
 
-class TestGitHub2GitLab(testtools.TestCase):
+class TestGitHub2GitLab(object):
 
-    def setUp(self):
-        super(TestGitHub2GitLab, self).setUp()
+    def setup(self):
         self.gitlab_url = 'http://gitlab'
         self.gitlab_token = 'token'
         self.github_repo = 'user/repo'
         self.g = main.GitHub2GitLab.factory([
+            '--verbose',
             '--gitlab-url', self.gitlab_url,
             '--gitlab-token', self.gitlab_token,
             '--github-repo', self.github_repo,
         ])
-        fixture = fixtures.TempDir()
-        self.useFixture(fixture)
-        self.d = fixture.path
+        self.d = tempfile.mkdtemp()
 
-    def tearDown(self):
-        super(TestGitHub2GitLab, self).tearDown()
+    def teardown(self):
+        shutil.rmtree(self.d)
 
     def test_init(self):
-        self.assertRaises(SystemExit, main.GitHub2GitLab.factory, [])
-        self.assertIn(os.environ['HOME'], self.g.args.ssh_public_key)
-        self.assertEqual(self.github_repo, self.g.github['repo'])
-        self.assertIn(self.gitlab_url, self.g.gitlab['url'])
+        with pytest.raises(SystemExit):
+            main.GitHub2GitLab.factory([])
+        assert os.environ['HOME'] in self.g.args.ssh_public_key
+        assert self.github_repo == self.g.github['repo']
+        assert self.gitlab_url in self.g.gitlab['url']
 
     @mock.patch('requests.get')
     def test_get(self, m_requests_get):
         g = self.g
-        class Request:
+
+        class Request(object):
             def __init__(self, params):
                 if params.get('page') == '1':
                     self.payload = [1]
@@ -77,38 +78,38 @@ class TestGitHub2GitLab(testtools.TestCase):
         m_requests_get.side_effect = lambda url, params: Request(params)
         result = self.g.get(self.g.gitlab['url'], {'key': 'value'},
                             cache=False)
-        m_requests_get.assert_called()
-        self.assertEqual([0, 1], result)
+        assert m_requests_get.called
+        assert [0, 1] == result
         other_result = self.g.get(self.g.gitlab['url'], {'key': 'value'},
                                   cache=False)
-        self.assertEqual(result, other_result)
+        assert result == other_result
 
     @mock.patch('requests.get')
     def test_get_pull_requests(self, m_requests_get):
         number1 = 1
         number2 = 2
 
-        class Request:
+        class Request(object):
             def __init__(self):
                 self.headers = {}
 
             def json(self):
-                return [{"number":number1},
-                        {"number":number2}]
+                return [{"number": number1},
+                        {"number": number2}]
 
         m_requests_get.side_effect = lambda url, params: Request()
         result = self.g.get_pull_requests()
-        self.assertEqual({
-            str(number1): { u'number': number1 },
-            str(number2): { u'number': number2 },
-        }, result)
+        assert {
+            str(number1): {u'number': number1},
+            str(number2): {u'number': number2},
+        } == result
 
     @mock.patch('requests.get')
     def test_get_merge_requests(self, m_requests_get):
         id1 = 100
         id2 = 200
 
-        class Request:
+        class Request(object):
             def __init__(self):
                 self.headers = {}
 
@@ -118,19 +119,20 @@ class TestGitHub2GitLab(testtools.TestCase):
 
         m_requests_get.side_effect = lambda url, params: Request()
         result = self.g.get_merge_requests()
-        self.assertEqual({
-            str(id1): { u'id': id1 },
-            str(id2): { u'id': id2 },
-        }, result)
+        assert {
+            str(id1): {u'id': id1},
+            str(id2): {u'id': id2},
+        } == result
 
     @mock.patch('requests.put')
     @mock.patch('requests.get')
     def test_unprotect_branches(self,
                                 m_requests_get,
                                 m_requests_put):
-        class Get:
+        class Get(object):
             def raise_for_status(self):
                 pass
+
             def json(self):
                 return [
                     {
@@ -143,13 +145,14 @@ class TestGitHub2GitLab(testtools.TestCase):
                     },
                 ]
         m_requests_get.side_effect = lambda url, params: Get()
-        class Put:
+
+        class Put(object):
             def raise_for_status(self):
                 pass
         m_requests_put.side_effect = lambda url, params: Put()
-        self.assertEquals(1, self.g.unprotect_branches())
-        m_requests_get.assert_called()
-        m_requests_put.assert_called()
+        assert 1 == self.g.unprotect_branches()
+        assert m_requests_get.called
+        assert m_requests_put.called
 
     def test_update_merge_pull(self):
         id1 = '100'
@@ -157,8 +160,8 @@ class TestGitHub2GitLab(testtools.TestCase):
         number1 = '1'
         number2 = '2'
         self.g.pull_requests = {
-            number1: { u'number': int(number1) },
-            number2: { u'number': int(number2) },
+            number1: {u'number': int(number1)},
+            number2: {u'number': int(number2)},
         }
         self.g.merge_requests = {
             id1: {
@@ -171,16 +174,15 @@ class TestGitHub2GitLab(testtools.TestCase):
             },
         }
         self.g.update_merge_pull()
-        self.assertEquals(self.g.merge2pull,
-                          {'100': {u'number': 1}})
-        self.assertEquals(self.g.pull2merge,
-                          {'1': {u'id': 100, u'source_branch': 'pull/1/head'}})
-
+        assert self.g.merge2pull == {'100': {u'number': 1}}
+        assert (self.g.pull2merge ==
+                {'1': {u'id': 100, u'source_branch': 'pull/1/head'}})
 
     @mock.patch('requests.post')
     def test_create_merge_request(self, m_requests_post):
-        data = { 'title': u'TITLE é' }
-        class Request:
+        data = {'title': u'TITLE é'}
+
+        class Request(object):
             def __init__(self):
                 self.status_code = 201
 
@@ -189,21 +191,21 @@ class TestGitHub2GitLab(testtools.TestCase):
 
         m_requests_post.side_effect = lambda url, params: Request()
         self.g.create_merge_request(data)
-        m_requests_post.assert_called()
-        
+        assert m_requests_post.called
+
     @mock.patch('requests.post')
     def test_create_merge_request_fail(self, m_requests_post):
-        data = { 'title': u'TITLE é' }
-        class Request:
-            def __init__(self, data):
-                self.data = data
+        data = {'title': u'TITLE é'}
 
-            def read(self):
-                return {}
+        class Request(object):
+            def __init__(self):
+                self.status_code = 400
+                self.text = 'FAIL'
 
-        m_requests_post.side_effect = lambda url, data: Request(data)
-        self.assertRaises(Exception, self.g.create_merge_request, data)
-        m_requests_post.assert_called()
+        m_requests_post.side_effect = lambda *args, **kwargs: Request()
+        with pytest.raises(ValueError):
+            self.g.create_merge_request(data)
+        assert m_requests_post.called
 
     @mock.patch('requests.put')
     def test_update_merge_request(self, m_requests_put):
@@ -211,20 +213,22 @@ class TestGitHub2GitLab(testtools.TestCase):
             'title': u'TITLE é',
             'state_event': 'close',
         }
-        class Request:
+
+        class Request(object):
             def json(self):
                 data['state'] = 'closed'
                 return data
 
         m_requests_put.side_effect = lambda url, params: Request()
-        merge_request = { 'id': 3 }
+        merge_request = {'id': 3}
         self.g.update_merge_request(merge_request, data)
-        m_requests_put.assert_called()
-        
+        assert m_requests_put.called
+
     @mock.patch('requests.put')
     def test_update_merge_request_fail_state(self, m_requests_put):
-        data = { 'state_event': 'close' }
-        class Request:
+        data = {'state_event': 'close'}
+
+        class Request(object):
             def json(self):
                 return {
                     'state': 'UNEXPECTED',
@@ -233,12 +237,11 @@ class TestGitHub2GitLab(testtools.TestCase):
                 }
 
         m_requests_put.side_effect = lambda url, params: Request()
-        merge_request = { 'id': 3 }
-        self.assertRaisesRegexp(Exception, 'UNEXPECTED',
-                                self.g.update_merge_request,
-                                merge_request,
-                                data)
-        m_requests_put.assert_called()
+        merge_request = {'id': 3}
+        with pytest.raises(ValueError) as e:
+            self.g.update_merge_request(merge_request, data)
+        assert 'UNEXPECTED' in str(e)
+        assert m_requests_put.called
 
     @mock.patch('github2gitlab.main.GitHub2GitLab.put_merge_request')
     @mock.patch('github2gitlab.main.GitHub2GitLab.verify_merge_update')
@@ -246,6 +249,7 @@ class TestGitHub2GitLab(testtools.TestCase):
                                         m_verify_merge_update,
                                         m_put_merge_request):
         description = 'DESCRIPTION'
+
         def put(merge_request, updates):
             if updates['state_event'] == 'merge':
                 updates['state'] = 'opened'
@@ -255,22 +259,24 @@ class TestGitHub2GitLab(testtools.TestCase):
             return updates
 
         m_put_merge_request.side_effect = put
-        merge_request = { 'id': 3 }
+        merge_request = {'id': 3}
         result = self.g.update_merge_request(merge_request,
-                                             { 'state_event': 'merge' })
-        self.assertIn(self.g.TAG_MERGED, result['description'])
-        m_verify_merge_update.assert_called()
-        
+                                             {'state_event': 'merge'})
+        assert self.g.TAG_MERGED in result['description']
+        assert m_verify_merge_update.called
+
     @mock.patch('requests.get')
     @mock.patch('requests.post')
     def test_add_project_create(self,
                                 m_requests_post,
                                 m_requests_get):
-        class Get:
+        class Get(object):
+
             def __init__(self):
                 self.status_code = 404
         m_requests_get.side_effect = lambda url, params: Get()
-        class Post:
+
+        class Post(object):
             def __init__(self):
                 self.status_code = 201
                 self.text = 'true'
@@ -278,38 +284,42 @@ class TestGitHub2GitLab(testtools.TestCase):
             def json(self):
                 return {}
         m_requests_post.side_effect = lambda url, params: Post()
-        self.assertEquals({}, self.g.add_project())
-        m_requests_get.assert_called()
-        m_requests_post.assert_called()
+        assert {} == self.g.add_project()
+        assert m_requests_get.called
+        assert m_requests_post.called
 
     @mock.patch('requests.get')
     @mock.patch('requests.post')
     def test_add_project_create_400(self,
                                     m_requests_post,
                                     m_requests_get):
-        class Get:
+        class Get(object):
+
             def __init__(self):
                 self.status_code = 404
         m_requests_get.side_effect = lambda url, params: Get()
         error_message = 'ERROR MESSAGE'
-        class Post:
+
+        class Post(object):
             def __init__(self):
                 self.status_code = 400
                 self.text = error_message
         m_requests_post.side_effect = lambda url, params: Post()
-        self.assertRaisesRegexp(ValueError, error_message, self.g.add_project)
-        m_requests_get.assert_called()
-        m_requests_post.assert_called()
+        with pytest.raises(ValueError) as e:
+            self.g.add_project()
+        assert error_message in str(e)
+        assert m_requests_get.called
+        assert m_requests_post.called
 
     @mock.patch('requests.get')
     def test_add_project_noop(self, m_requests_get):
-        class Get:
+        class Get(object):
             def __init__(self):
                 self.status_code = 200
         m_requests_get.side_effect = lambda url, params: Get()
-        self.assertEquals(None, self.g.add_project())
-        m_requests_get.assert_called()
-        
+        assert None == self.g.add_project()
+        assert m_requests_get.called
+
     @mock.patch('requests.get')
     @mock.patch('requests.post')
     def test_add_key_create(self,
@@ -320,18 +330,20 @@ class TestGitHub2GitLab(testtools.TestCase):
         with open(ssh_public_key, 'w') as f:
             f.write(public_key)
         self.g.args.ssh_public_key = ssh_public_key
-        class Get:
+
+        class Get(object):
             def json(self):
                 return []
         m_requests_get.side_effect = lambda url, params: Get()
-        class Post:
+
+        class Post(object):
             def __init__(self):
                 self.status_code = 201
         m_requests_post.side_effect = lambda url, params: Post()
-        self.assertEquals(public_key, self.g.add_key())
-        m_requests_get.assert_called()
-        m_requests_post.assert_called()
-        
+        assert public_key == self.g.add_key()
+        assert m_requests_get.called
+        assert m_requests_post.called
+
     @mock.patch('requests.get')
     @mock.patch('requests.post')
     def test_add_key_create_400(self,
@@ -342,20 +354,24 @@ class TestGitHub2GitLab(testtools.TestCase):
         with open(ssh_public_key, 'w') as f:
             f.write(public_key)
         self.g.args.ssh_public_key = ssh_public_key
-        class Get:
+
+        class Get(object):
             def json(self):
                 return []
         m_requests_get.side_effect = lambda url, params: Get()
         error_message = 'ERROR MESSAGE'
-        class Post:
+
+        class Post(object):
             def __init__(self):
                 self.status_code = 400
                 self.text = error_message
         m_requests_post.side_effect = lambda url, params: Post()
-        self.assertRaisesRegexp(ValueError, error_message, self.g.add_key)
-        m_requests_get.assert_called()
-        m_requests_post.assert_called()
-        
+        with pytest.raises(ValueError) as e:
+            self.g.add_key()
+        assert error_message in str(e)
+        assert m_requests_get.called
+        assert m_requests_post.called
+
     @mock.patch('requests.get')
     def test_add_key_noop(self, m_requests_get):
         public_key = 'PUBLIC KEY'
@@ -363,14 +379,14 @@ class TestGitHub2GitLab(testtools.TestCase):
         with open(ssh_public_key, 'w') as f:
             f.write(public_key)
         self.g.args.ssh_public_key = ssh_public_key
-        class Get:
+
+        class Get(object):
             def json(self):
                 return [{'key': public_key}]
         m_requests_get.side_effect = lambda url, params: Get()
-        self.assertEquals(None, self.g.add_key())
-        m_requests_get.assert_called()
-        
-        
+        assert None == self.g.add_key()
+        assert m_requests_get.called
+
     @mock.patch('github2gitlab.main.GitHub2GitLab.update_merge_request')
     @mock.patch('github2gitlab.main.GitHub2GitLab.create_merge_request')
     @mock.patch('github2gitlab.main.GitHub2GitLab.rev_parse')
@@ -409,16 +425,17 @@ class TestGitHub2GitLab(testtools.TestCase):
             }
         }
         self.g.update_merge_pull()
-        self.assertEquals(1, len(self.g.merge2pull))
-        self.assertEquals(2, self.g.merge2pull['100']['number'])
-        self.assertEquals(100, self.g.pull2merge['2']['id'])
+        assert 1 == len(self.g.merge2pull)
+        assert 2 == self.g.merge2pull['100']['number']
+        assert 100 == self.g.pull2merge['2']['id']
 
         self.g.sync()
-        m_update_merge_request.assert_called_with(self.g.merge_requests['100'],
-                                                  {
-                                                      'title': 'OTHER_TITLE',
-                                                      'state_event': 'merge',
-                                                  })
+        m_update_merge_request.assert_called_with(
+            self.g.merge_requests['100'],
+            {
+                'title': 'OTHER_TITLE',
+                'state_event': 'merge',
+            })
         pull = self.g.pull_requests['1']
         m_create_merge_request.assert_called_with({
             'title': pull['title'],
@@ -426,14 +443,228 @@ class TestGitHub2GitLab(testtools.TestCase):
             'target_branch': pull['base']['ref'],
             'source_branch': 'pull/' + str(pull['number']) + '/head',
         })
-    
-class TestGitHub2GitLabNoSetup(testtools.TestCase):
+
+    @mock.patch('github2gitlab.main.GitHub2GitLab.gitlab_create_remote')
+    def test_gitmirror(self, m_gitlab_create_remote):
+        self.g.args.skip_pull_requests = True
+
+        self.g.sh("""
+        cd {dir}
+        mkdir github
+        cd github
+        git init
+        echo a > a ; git add a ; git commit -m "a" a
+        """.format(dir=self.d))
+
+        self.g.sh("""
+        cd {dir}
+        mkdir gitlab
+        cd gitlab
+        git init --bare
+        """.format(dir=self.d))
+
+        def gitlab_create_remote(repo):
+            repo.create_remote('gitlab', self.d + "/gitlab")
+        m_gitlab_create_remote.side_effect = gitlab_create_remote
+
+        self.g.github['git'] = self.d
+        self.g.github['repo'] = 'github'
+        self.g.gitlab['name'] = 'project'
+
+        cwd = os.getcwd()
+        os.chdir(self.d)
+        gitlab = git.Repo(self.d + '/gitlab')
+        github = git.Repo(self.d + '/github')
+
+        #
+        # A pull request newly created only has pull/1/head
+        # and will not be pushed to gitlab.
+        #
+        self.g.sh("""
+        cd {dir}/github
+        git branch pr-1
+        git checkout pr-1
+        echo 1 > 1 ; git add 1 ; git commit -m "1" 1
+        git update-ref refs/pull/1/head HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+
+        with pytest.raises(gitdb.exc.BadName):
+            gitlab.commit('pull/1/merge')
+
+        #
+        # After github successfully test a merge of a pull
+        # request, the pull/1/merge reference exists and
+        # it is pushed to gitlab
+        #
+        self.g.sh("""
+        cd {dir}/github
+        git branch pr-1-merge master
+        git checkout pr-1-merge
+        git merge --no-ff pr-1
+        git update-ref refs/pull/1/merge HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        assert gitlab.commit('pull/1/merge') == github.commit('pull/1/merge')
+        assert gitlab.commit('pull/1/head') == github.commit('pull/1/head')
+
+        #
+        # When the base (master in this case) changes, github
+        # may decide to re-try a merge. We ignore that because we're
+        # only interested in pushing to gitlab when the head changes,
+        # not every time the base changes and a merge is attempted.
+        #
+        self.g.sh("""
+        cd {dir}/github
+        git checkout master
+        echo b > b ; git add b ; git commit -m "b" b
+        git checkout pr-1-merge
+        git reset --hard master
+        git merge --no-ff pr-1
+        git update-ref refs/pull/1/merge HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        assert gitlab.commit('pull/1/merge') != github.commit('pull/1/merge')
+        assert gitlab.commit('pull/1/head') == github.commit('pull/1/head')
+
+        #
+        # After a rebase and repush, pull/2/merge may reference the
+        # previous pull/2/head and there is no point in pushing
+        # it to gitlab: it would show an outdated state of the
+        # pull request
+        #
+        self.g.sh("""
+        cd {dir}/github
+        # new pr
+        git branch pr-2
+        git checkout pr-2
+        echo 2 > 2 ; git add 2 ; git commit -m "2" 2
+        git update-ref refs/pull/2/head HEAD
+        git checkout master
+        # merge test
+        git branch pr-2-merge master
+        git checkout pr-2-merge
+        git merge --no-ff pr-2
+        git update-ref refs/pull/2/merge HEAD
+        git checkout master
+        # repush the pr
+        git checkout pr-2
+        echo 2.5 > 2.5 ; git add 2.5 ; git commit -m "2.5" 2.5
+        git update-ref refs/pull/2/head HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        with pytest.raises(gitdb.exc.BadName):
+            gitlab.commit('pull/2/merge')
+
+        #
+        # After a rebase and repush, pull/3/merge is updated
+        # if the merge check is successful.
+        #
+        self.g.sh("""
+        cd {dir}/github
+        # new pr
+        git branch pr-3
+        git checkout pr-3
+        echo 3 > 3 ; git add 3 ; git commit -m "3" 3
+        git update-ref refs/pull/3/head HEAD
+        git checkout master
+        # merge test
+        git branch pr-3-merge master
+        git checkout pr-3-merge
+        git merge --no-ff pr-3
+        git update-ref refs/pull/3/merge HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        merge_1 = gitlab.commit('pull/3/merge')
+        assert merge_1 == github.commit('pull/3/merge')
+        head_1 = gitlab.commit('pull/3/head')
+        assert head_1 == github.commit('pull/3/head')
+
+        self.g.sh("""
+        cd {dir}/github
+        # repush the pr
+        git checkout pr-3
+        echo 3.5 > 3.5 ; git add 3.5 ; git commit -m "3.5" 3.5
+        git update-ref refs/pull/3/head HEAD
+        git checkout master
+        # re-verify merge
+        git checkout pr-3-merge
+        git reset --hard master
+        git merge --no-ff pr-3
+        git update-ref refs/pull/3/merge HEAD
+        git checkout master
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        merge_2 = gitlab.commit('pull/3/merge')
+        assert merge_1 != merge_2
+        assert merge_2 == github.commit('pull/3/merge')
+        head_2 = gitlab.commit('pull/3/head')
+        assert head_1 != head_2
+        assert head_2 == github.commit('pull/3/head')
+
+        #
+        # All branches are pushed to gitlab by default
+        #
+        self.g.sh("""
+        cd {dir}/github
+        git branch b master
+        git checkout b
+        echo b > b ; git add b ; git commit -m "b" b
+        git branch c master
+        git checkout c
+        echo c > c ; git add c ; git commit -m "c" c
+        """.format(dir=self.d))
+
+        self.g.git_mirror()
+        assert gitlab.commit('b') == github.commit('b')
+        assert gitlab.commit('c') == github.commit('c')
+
+        #
+        # If branches are given in argument, only those
+        # are pushed to gitlab
+        #
+        self.g.sh("""
+        cd {dir}/github
+        git branch d master
+        git checkout d
+        echo d > d ; git add d ; git commit -m "d" d
+        git branch e master
+        git checkout e
+        echo e > e ; git add e ; git commit -m "e" e
+        git branch f master
+        git checkout f
+        echo f > f ; git add f ; git commit -m "f" f
+        """.format(dir=self.d))
+
+        self.g.github['branches'] = ['e', 'f']
+        self.g.git_mirror()
+        with pytest.raises(gitdb.exc.BadName):
+            gitlab.commit('d')
+        assert gitlab.commit('e') == github.commit('e')
+        assert gitlab.commit('f') == github.commit('f')
+
+        os.chdir(cwd)
+
+
+class TestGitHub2GitLabNoSetup(object):
 
     def test_json_loads(self):
         r = main.GitHub2GitLab.json_loads('{}')
-        self.assertEquals({}, r)
-        self.assertRaises(ValueError, main.GitHub2GitLab.json_loads, ']')
+        assert {} == r
+        with pytest.raises(ValueError):
+            main.GitHub2GitLab.json_loads(']')
 
 # Local Variables:
-# compile-command: "../.tox/py33/bin/py.test test_github2gitlab.py"
+# compile-command: "../.tox/py27/bin/py.test test_github2gitlab.py"
 # End:
